#!/usr/bin/env python
import datetime
import fcntl
import getpass
import glob
import json
import os
import pexpect
import requests
import re
import shutil
import subprocess
import sys
import tempfile
import time

LOGFILE = 'log50.log'
LOGDIR = '/var/log/log50/'
FILESDIR = '/var/log/log50/files/'
SERVERFILESDIR = 'files/'
UPLOADDIR = '/var/log/log50/'
NO50_PATH = '/etc/no50'
APPLIANCE_ID_FILE = '/etc/appliance50.id'
LOCKFILE = '/var/lock/log50'
METADATAFILE = '.uploaded'
DEBUG = False
LOGURL = 'https://log19.cs50.net'
UPLOADHOURS = 24 # number of hours after which to reupload
MAX_CONTENT_SIZE = 1024 * 1024 # max number of bytes of content to log for scripts - 1MB, currently

"""
Runs a program, recording a script of execution in the log
"""
class ScriptWrapper:
    def __init__(self, progName, extraData = None, realProgPath = None):
        # if no real path is found, find it with which
        if realProgPath is None:
            realProgPath = self.findRealProgPath(progName)
            if realProgPath is None:
                print "%s: command not found" % progName
                exit()

        # set instance variables
        self.progName = progName
        self.realProgPath = realProgPath
        self.extraData = extraData

    def run(self):
        # gen rand file name for pexpect logfile
        temp = tempfile.NamedTemporaryFile(delete=False)
        os.chown(temp.name, os.geteuid(), os.getegid())

        # run program, modifying argv to go to real program
        sys.argv[0] = self.realProgPath

        # build command
        command = ' '.join(sys.argv)

        start_time = time.time()
        p = pexpect.spawn(command)
        p.logfile = temp

        # removed because unnecessary
        #global global_pexpect_instance
        #global_pexpect_instance = p
        #signal.signal(signal.SIGWINCH, sigwinch_passthrough)

        try:
            p.interact()

        # for I/O problems on quit
        except OSError:
            pass

        end_time = time.time()

        # read in logfile
        f = open(temp.name)
        content = f.read()
        temp.close()

        p.close()

        # cap size of content
        content = content[:MAX_CONTENT_SIZE]

        # create object to be logged
        obj = {'_type':'wrapper', 'prog': self.progName, 'cmd': command, 'exit_code': p.exitstatus, 'start_time':start_time, 'end_time':end_time, 'content':content, 'cwd':os.getcwd(), 'username': getpass.getuser()}

        # add extra data if provided
        if self.extraData is not None:
            obj.update(self.extraData)

        # write log
        log_obj(obj)


    def findRealProgPath(self, progName):
        try:
            # call which on program name
            output = subprocess.check_output(['which', '-a', progName])
            lines = output.split('\n')

            # if only one result, fail
            if len(lines) < 2:
                return None
            else:
                cmd = lines[0]
                # if log50.d is in path multiple times, we might need to index for a bit
                index = 1
                # increase until new path found (if we run out, return None)
                while lines[index-1] == lines[index] or lines[index] == '' or lines[index] == cmd:
                    index = index + 1
                    if index > len(lines)-1:
                        return None

                # if new path found, return it
                return lines[index]

        # if call failed, fail
        except:
            return None

def can_log():
    # if no50 exists, pass
    return not os.path.exists(NO50_PATH)

def log_obj(obj):
    if not can_log():
        return

    # add timestamp and turn to string
    obj['_timestamp'] = time.time()
    string = json.dumps(obj)

    try:
        with open(LOGDIR + LOGFILE, "a+") as logfile:
            # lock file. lock should be released when file is closed
            # XXX: should this be nonblocking?
            fcntl.flock(logfile, fcntl.LOCK_EX)

            try:
                logfile.write("%s\n" % string)
            except:
                pass
    except Exception, e:
        if DEBUG:
            print "failed to log object"
            print e
        pass

# convenience method for logging strings
def log_str(string):
    obj = {"message": string}
    log_obj(obj)

"""
Saves a copy of a file to the log directory, diffing if possible and diff is True
Returns the path to the file in the log folder
"""
def log_file(path, diff=False):
    if not can_log():
        return

    # acquire a lock, so that only one file is being copied at a time
    lock = acquire_lock(True)
    if lock is None:
        if DEBUG:
            print "Failed to acquire file lock"
            return None

    # construct path for file (removing leading slash from path)
    # XXX: make sure trailing slash is only there once
    if path[0] == '/':
        log_path = FILESDIR + path[1:] + '/'
    else:
        log_path = FILESDIR + path + '/'

    # check if directory exists (no need to worry about race conditions,
    # because only one file copy is allowed at once
    #XXX: need to check if dir?
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    # add file to directory. as diff if diff is specified, otherwise whole file
    if diff:
        # if there is not a "current" file, just record a new base
        if not os.path.isfile(log_path + '.current'):
            return make_base(path, log_path)

        # else, just diff with the current file, to save space
        else:
            return make_diff(path, log_path + '.current', log_path,)

    else:
        return make_base(path, log_path)

"""
Makes a copy of 'path' as a new base copy in the 'log_path' directory
Returns path to new file, or None on error.
"""
def make_base(path, log_path):
    # copy logfile with time suffix
    copy_time = time.time()

    # create new file name
    dest = log_path + 'base.' + str(copy_time)
    # copy in with new name (TODO: error check on name)
    try:
        shutil.copy(path, dest)
        shutil.copy(dest, log_path + '.current')

    except Exception, e:
        if DEBUG:
            print "failed to copy file"
            print e
        return None

    # return dest, if all suceeeded
    return dest

"""
Makes an ed script from 'old' to 'new' as a new file in the 'log_path' directory
Returns path to diff file, or None on error.
"""
def make_diff(new, old, log_path):
    # get current time to add onto end of diff file
    copy_time = time.time()

    # create path for destination file
    dest = log_path + 'diff.' + str(copy_time)

    # create diff
    try:
        with open(dest, 'w') as diff_file:
            res = subprocess.call(['diff', '-e', old, new], stdout=diff_file)

            # if no diff (res == 0), file will not have been created, so create a blank diff file
            if res == 0:
                diff_file.write('')

        # update current file
        shutil.copy(new, log_path + '.current')

    except Exception, e:
        if DEBUG:
            print "failed to create diff of %s" % new
            print e
        return None

    # on sucess, return the diff path
    return dest


"""
Copies log into log directory, with timestamp in filename.
If sucessful, wipes log file.
"""
def copy_log(log_file=LOGFILE, log_dir=LOGDIR, dest_dir=UPLOADDIR):
    # open and lock logfile
    log_file_path = log_dir + log_file
    fd = open(log_file_path,"r")

    # try to lock file
    try:
        fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except Exception, e:
        if DEBUG:
            print "Failed to lock log file"
            print e
        fd.close()
        return False

    # initialize success
    success = None

    try:
        # create upload dir if necessary
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir, 0755)

        # copy logfile with time suffix
        copy_time = time.time()

        # copy in with new name (TODO: error check on name)
        shutil.copy(log_dir+log_file, dest_dir + log_file + '.' + str(copy_time))

        # zeroes out the log. TODO necessary to lock first?
        open(log_file_path,"w").close()

        success = True
    except Exception, e:
        if DEBUG:
            print "Failed to copy log"
            print e
        success = False

    # unlock file and return
    fd.close()
    return success


"""
Uploads one file to the logging site, and optionally deletes it afterwards
"""
def send_file(path, name, appliance_id, log_url, delete=False):
    # if debugging, print file names
    if DEBUG:
        print "attempting to upload %s as %s" %(path, name)

    file = open(path, 'rb')

    # make POST request once per file, continuing on error
    # log.cs50.net will use the key of the FILES array, unless a name is specified
    try:
        r = requests.post(log_url, data={'appliance_id':appliance_id, 'name':name}, files={name:file})
    except:
        if DEBUG:
            print "Request failed for %s." % path
        return False

    # validate response
    if r is not None and r.status_code == 200:
        try:
            response = json.loads(r.text)
        except ValueError:
            if DEBUG:
                print "could not parse ", r.content
            return False

        # see if response succeded
        if "success" in response and response["success"]:
            # unlink file and continue
            if delete:
                os.unlink(path)
            return True
        else:
            if DEBUG:
                print "Request failed for %s.\nError:%s" % (path, r.content)
            return False
    else:
        if DEBUG:
            print "Request failed for %s.\nError:%s" % (path, r.content)
        return False

"""
Send files located in files_path to the server. It uploads them into the server_files_dir folder on the server
"""
def send_files(appliance_id, files_dir=FILESDIR, server_files_dir = SERVERFILESDIR, log_url = LOGURL):
    # get data files to log as well
    for path, dirs, files in os.walk(files_dir):
        # (note: this retains the '/files/' portion of the name
        file_dir_len = len(files_dir)
        for file in files:
            # add to upload files
            full_path = os.path.join(path, file)
            name = server_files_dir + full_path[file_dir_len:]

            send_file(full_path, name, appliance_id, log_url, True)

def send_log(appliance_id, log_dir=UPLOADDIR, log_url = LOGURL):
    if not can_log():
        return

    # find all files in log directory that end in one or more digits
    log_files = glob.glob(log_dir + "*.log.[0-9][0-9]*")

    if len(log_files) > 0:
        # upload one file at a time
        prefix_len = len(log_dir)
        for file_name in log_files:
            # get name for path on server
            name = file_name[prefix_len:]
            send_file(file_name, name, appliance_id, log_url, True)


"""
Tries to acquire the lock.
Returns the locked file descriptor if is lock acquired, otherwise None.
"""
def acquire_lock(block):
    try:
        # open lock file for editing
        # NOTE: if lockfile does not exist, this will create it and return false, due to need to loc
        lockfile = open(LOCKFILE, "a+")
        # lock file. lock should be released when file is closed
        # if lock file is currently locked, return false
        try:
            if block:
                fcntl.flock(lockfile, fcntl.LOCK_EX)
            else:
                fcntl.flock(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except Exception, e:
            if DEBUG:
                print "Failed to acquire lock"
                print e
            return None

        # if succeeded in acquiring lock, return true
        return lockfile

    # if error occurs, return false
    except Exception, e:
        if DEBUG:
            print "Failed to acquire lock"
            print e
        return None

"""
Unlocks the lock file
"""
def release_lock(lock):
    # unlock file
    try:
        fcntl.flock(lock, fcntl.LOCK_UN)
    except Exception, e:
        if DEBUG:
            print "Failed to release lock"
            print e


def should_upload():
    md_file_path = LOGDIR+METADATAFILE

    # if file doesn't exist, return true
    if not os.path.exists(md_file_path):
        return True

    # else check if file is new enough
    else:
        modified_time = os.path.getmtime(md_file_path)
        now = time.time()

        # get difference in hours
        hours_since = (now - modified_time) / 3600

        # return true if its been over 24 hours, or if something went wrong and the time is negative
        return hours_since < 0 or hours_since > UPLOADHOURS

def get_appliance_id():
    try:
        output = subprocess.check_output(['id50'])
        return output.rstrip()
    except:
        if DEBUG:
            print "no appliance id found"
        return None
